# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import shutil
import os
from glob import glob

project = 'ASSYST'
copyright = '2025, Max-Planck-Institute for Sustainable Materials'
author = 'Marvin Poul'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinx.ext.intersphinx',
    'myst_nb',
]

nb_execution_mode = "off"

# Copy notebooks from root to docs/notebooks
# This ensures that notebooks are available to sphinx regardless of where the build is run
# and avoids having to hardcode them in index.rst
nb_source_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'notebooks'))
nb_dest_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), 'notebooks'))

if os.path.exists(nb_source_dir):
    if not os.path.exists(nb_dest_dir):
        os.makedirs(nb_dest_dir)
    for file in glob("**/*.ipynb", root_dir=nb_source_dir, recursive=True):
        dest = os.path.join(nb_dest_dir, file)
        os.makedirs(os.path.dirname(dest), exist_ok=True)
        shutil.copy(os.path.join(nb_source_dir, file), dest)

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'matplotlib': ('https://matplotlib.org/stable/', None),
    'ase': ('https://wiki.fysik.dtu.dk/ase', None),
}

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'notebooks/.*']


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'furo'
html_static_path = ['_static']



"""Utility module to pick where to load neighbour_list from."""

from pyiron_snippets.import_alarm import ImportAlarm

with ImportAlarm(
    "matscipy not installed falling back to ASE implementation; "
    "install with 'conda install -c conda-forge matscipy' or "
    "'pip install matscipy'",
) as neighbor_alarm:
    try:
        from matscipy.neighbours import neighbour_list as neighbor_list
    except ImportError:
        from ase.neighborlist import neighbor_list

        raise

__all__ = ["neighbor_list"]



import pytest
from ase import Atoms
from assyst.crystals import pyxtal


def test_num_ions_and_species_length_mismatch_one():
    """ValueError must be raised if num_ions and species lengths should not match"""
    with pytest.raises(ValueError):
        pyxtal(1, species=["Fe"], num_ions=[1, 2])

    with pytest.raises(ValueError):
        pyxtal(1, species=["Fe", "Cr"], num_ions=[1])


def test_allow_exceptions_true_warns():
    """allow_exceptions=True should suppress the error and emit a warning!"""
    try:
        with pytest.warns(
            UserWarning,
            match=r"Groups \[193, 194\] could not be generated with stoichiometry Mg1!",
        ) as record:
            pyxtal([193, 194], ["Mg"], num_ions=[1], allow_exceptions=True)

        assert len(record) == 1, "warning count should be exactly 1!"
    except ValueError:
        pytest.fail("Error should not be raised when allow_exceptions=True is passed!")


def test_allow_exceptions_false_raises():
    """allow_exceptions=False should let the ValueError propagate!"""
    with pytest.raises(ValueError):
        pyxtal(194, ["Mg"], num_ions=[1], allow_exceptions=False)


def test_return_is_atoms_instance():
    """pyxtal should return an Atoms instance for scalar arguments!"""
    result = pyxtal(1, species=["Fe"], num_ions=[1])
    assert isinstance(result, Atoms), "result should be an Atoms instance!"


def test_return_is_list_for_multiple_groups():
    """pyxtal should return a list when multiple space‑group numbers are supplied!"""
    result = pyxtal([1, 2], species=["Fe"], num_ions=[1])
    assert isinstance(result, list), "result should be a list when multiple groups are supplied!"


def test_return_is_list_when_repeat_given():
    """pyxtal should return a list when the repeat argument is used!"""
    result = pyxtal(1, species=["Fe"], num_ions=[1], repeat=5)
    assert isinstance(result, list), "result should be a list when repeat is given!"


def test_repeat_length_matches_requested():
    """pyxtal should produce exactly the requested number of structures when repeat is set!"""
    result = pyxtal(1, species=["Fe"], num_ions=[1], repeat=5)
    assert len(result) == 5, "length of result should match the repeat value!"


def test_repeat_elements_are_dicts():
    """pyxtal should return a list of dicts when repeat is used!"""
    result = pyxtal(1, species=["Fe"], num_ions=[1], repeat=5)
    assert all(isinstance(d, dict) for d in result), "every element in result should be a dict!"



import unittest
from ase import Atoms
from assyst.filters import AndFilter, OrFilter, FilterBase

class MockFilter(FilterBase):
    def __init__(self, value):
        self.value = value
    def __call__(self, structure: Atoms) -> bool:
        return self.value

class TestComposition(unittest.TestCase):
    def test_and_filter(self):
        structure = Atoms()
        # True and True
        self.assertTrue(AndFilter(MockFilter(True), MockFilter(True))(structure))
        # True and False
        self.assertFalse(AndFilter(MockFilter(True), MockFilter(False))(structure))
        # False and True
        self.assertFalse(AndFilter(MockFilter(False), MockFilter(True))(structure))
        # False and False
        self.assertFalse(AndFilter(MockFilter(False), MockFilter(False))(structure))

    def test_or_filter(self):
        structure = Atoms()
        self.assertTrue(OrFilter(MockFilter(True), MockFilter(True))(structure))
        self.assertTrue(OrFilter(MockFilter(True), MockFilter(False))(structure))
        self.assertTrue(OrFilter(MockFilter(False), MockFilter(True))(structure))
        self.assertFalse(OrFilter(MockFilter(False), MockFilter(False))(structure))

    def test_and_operator(self):
        structure = Atoms()
        and_filter = MockFilter(True) & MockFilter(False)
        self.assertIsInstance(and_filter, AndFilter)
        self.assertFalse(and_filter(structure))

    def test_or_operator(self):
        structure = Atoms()
        or_filter = MockFilter(True) | MockFilter(False)
        self.assertIsInstance(or_filter, OrFilter)
        self.assertTrue(or_filter(structure))



"""Classes that filter structures according to some criteria.

The code in the other modules that uses them is set up such that simple
functions can always be passed as well and that the classes here are just for
convenience."""

from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, KW_ONLY
from itertools import combinations_with_replacement, product
from math import nan, inf
from typing import Callable, Literal

from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator
from pyxtal.tolerance import Tol_matrix
from ase.data import atomic_numbers
import numpy as np

from assyst.neighbors import neighbor_list


class FilterBase(ABC):
    """Base class for filter objects that implements conjunction and disjunction operators."""

    def __and__(self, other) -> "AndFilter":
        return AndFilter(self, other)

    def __or__(self, other) -> "OrFilter":
        return OrFilter(self, other)

    @abstractmethod
    def __call__(self, structure: Atoms) -> bool:
        """Returns True if structure passes the filter, False if it should be dropped."""
        pass


Filter = Callable[[Atoms], bool] | FilterBase


@dataclass(frozen=True, eq=True)
class AndFilter(FilterBase):
    """Conjunction of two filters."""

    l: Filter
    r: Filter

    def __call__(self, structure: Atoms) -> bool:
        return self.l(structure) and self.r(structure)


@dataclass(frozen=True, eq=True)
class OrFilter(FilterBase):
    """Disjunction of two filters."""

    l: Filter
    r: Filter

    def __call__(self, structure: Atoms) -> bool:
        return self.l(structure) or self.r(structure)


@dataclass
class DistanceFilter(FilterBase):
    """Filter structures that contain too close atoms.

    Setting a radius to NaN allows all bonds involving this atom."""

    radii: dict[str, float]

    @staticmethod
    def _element_wise_dist(structure: Atoms) -> dict[tuple[str, str], float]:
        pair: dict[tuple[str, str], float] = defaultdict(lambda: inf)
        for i, j, d in zip(*neighbor_list("ijd", structure, 5.0)):
            ei, ej = sorted((structure.symbols[i], structure.symbols[j]))
            pair[ei, ej] = min(d, pair[ei, ej])
        return pair

    def __call__(self, structure: Atoms) -> bool:
        """
        Return True if structure satifies minimum distance criteria.

        Args:
            structure (:class:`ase.Atoms`): structure to check

        Returns:
            `False`: at least on bond is shorter than the sum of given cutoff radii of the respective elements
            `True`: all bonds are than the sum of given cutoff radii of the respective elements
        """
        pair = self._element_wise_dist(structure)
        for ei, ej in combinations_with_replacement(structure.symbols.species(), 2):
            ei, ej = sorted((ei, ej))
            if pair[ei, ej] < self.radii.get(ei, nan) + self.radii.get(ej, nan):
                return False
        return True

    def to_tol_matrix(
        self, prototype: Literal["metallic", "atomic", "molecular", "vdW"] = "metallic"
    ) -> Tol_matrix:
        """Returns equivalent tolerance matrix for pyxtal.

        Args:
            prototype (metallic, atomic, molecular or vdW):
                passed to Tol_matrix as is and used there to initialize radii of elements not explicitly set in this
                filter
        """
        return Tol_matrix(
            *(
                (
                    atomic_numbers[e1],
                    atomic_numbers[e2],
                    self.radii[e1] + self.radii[e2],
                )
                for e1, e2 in product(self.radii, repeat=2)
            ),
            prototype=prototype,
        )


@dataclass
class AspectFilter(FilterBase):
    """Filters structures with high aspect ratios."""

    maximum_aspect_ratio: float = 6

    def __call__(self, structure: Atoms) -> bool:
        """Return True if structure's cell has an agreeable aspect ratio.

        Args:
            structure (:class:`ase.Atoms`): structure to check

        Returns:
            `True`: lattice's aspect ratio is below or equal `:attr:`.maximum_aspect_ratio`.
            `False`: lattice's aspect ratio is above `:attr:`.maximum_aspect_ratio`."""
        a, b, c = sorted(structure.cell.lengths())
        return c / a <= self.maximum_aspect_ratio


@dataclass
class VolumeFilter(FilterBase):
    """Filters structures by volume."""

    maximum_volume_per_atom: float

    def __call__(self, structure: Atoms) -> bool:
        """Return True if structure's volume is within range.

        Args:
            structure (:class:`ase.Atoms`): structure to check

        Returns:
            `True`: volume per atom is smaller or equal than `:attr:.maximum_volume_per_atom`.
            `False`: otherwise"""
        return structure.cell.volume / len(structure) <= self.maximum_volume_per_atom


@dataclass
class CalculatorFilter(FilterBase):
    """Filters that require a single point calculator set on the structure."""

    _: KW_ONLY
    missing: Literal["error", "ignore"] = "error"
    """What to do when a structure has no (correct) calculator attached."""

    def _check(self, structure: Atoms) -> bool:
        match self.missing:
            case "error":
                if structure.calc is None:
                    raise ValueError("Structure must have single point calculator set!")
                if not isinstance(structure.calc, SinglePointCalculator):
                    raise ValueError(
                        f"Structure must have single point calculator set, not {type(structure.calc)}!"
                    )
                return True
            case "ignore":
                return False
            case _:
                assert False


@dataclass
class EnergyFilter(CalculatorFilter):
    """Filters structures by energy per atom."""

    min_energy: float = -inf
    max_energy: float = +inf

    def __call__(self, structure: Atoms) -> bool:
        if not self._check(structure):
            return True
        return (
            self.min_energy
            <= structure.get_potential_energy() / len(structure)
            <= self.max_energy
        )


@dataclass
class ForceFilter(CalculatorFilter):
    """Filters structures by maximum force magnitude."""

    max_force: float = +inf

    def __call__(self, structure: Atoms) -> bool:
        if not self._check(structure):
            return True
        return np.linalg.norm(structure.get_forces(), axis=-1).max() <= self.max_force


__all__ = [
        "FilterBase",
        "Filter",
        "AndFilter",
        "OrFilter",
        "DistanceFilter",
        "AspectFilter",
        "VolumeFilter",
        "CalculatorFilter",
        "EnergyFilter",
        "ForceFilter"
]



"""
Convenience shorts to create ASE calculators to be used inside ASSYST.

Exists mostly to avoid passing around potentially large and unpickle-able calculator objects.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass, asdict
from functools import lru_cache
import os

from ase.calculators.calculator import Calculator
from ase.calculators.morse import MorsePotential
from pyiron_snippets.import_alarm import ImportAlarm

with ImportAlarm(
    "grace-tensorpotential required; install with "
    "'conda install -c conda-forge grace-tensorpotential' or 'pip install tensorpotential'",
    raise_exception=True,
) as grace_alarm:
    from tensorpotential.calculator import grace_fm


class AseCalculatorConfig(ABC):
    """Base class to keep calculator configurations."""

    @abstractmethod
    def get_calculator(self) -> Calculator:
        """Return the actual calculator object.

        Returns:
            :class:`ase.calculators.calculator.Calculator`: the actually usable calculator
        """
        pass


@dataclass(frozen=True, eq=True)
class Grace(AseCalculatorConfig):
    """Universal Graph Atomic Cluster Expansion models.

    .. attention::
        This class needs additional dependencies!
        Install `tensorpotential` from `PyPI <https://pypi.org/project/tensorpotential/>`__.
    """

    model: str = "GRACE-FS-OAM"

    @lru_cache(maxsize=1)
    @grace_alarm
    def get_calculator(self) -> Calculator:
        # disable tensorflow warnings noise
        os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
        return grace_fm(self.model)


@dataclass(frozen=True, eq=True)
class Morse(AseCalculatorConfig):
    """Morse potential for testing.  Parameters as in ASE."""

    epsilon: float = 1.0
    r0: float = 1.0
    rho0: float = 1.0

    def get_calculator(self) -> Calculator:
        return MorsePotential(**asdict(self))


__all__ = [
        "AseCalculatorConfig",
        "Grace",
        "Morse",
]



import unittest
from unittest.mock import patch
from math import inf

import pytest
from hypothesis import given, strategies as st
import numpy as np
from ase import Atoms
from ase.build import bulk
from ase.data import atomic_numbers

from assyst.perturbations import rattle, element_scaled_rattle, stretch, Rattle, Stretch, Series, RandomChoice, apply_perturbations


class TestPerturbations(unittest.TestCase):

    def setUp(self):
        """Create a simple Atoms object for testing."""
        self.structure = Atoms('H2', positions=[[0, 0, 0], [0.74, 0, 0]], cell=[10, 10, 10])
        self.single_atom_structure = Atoms('H', positions=[[0, 0, 0]], cell=[10, 10, 10])

    def test_rattle_modifies_positions(self):
        """Test that rattle modifies the positions of the atoms."""
        original_positions = self.structure.get_positions().copy()
        rattled_structure = rattle(self.structure.copy(), sigma=0.1)
        rattled_positions = rattled_structure.get_positions()
        self.assertFalse(np.allclose(original_positions, rattled_positions))

    def test_rattle_raises_value_error_for_single_atom(self):
        """Test that rattle raises a ValueError for a single-atom structure."""
        with self.assertRaises(ValueError):
            rattle(self.single_atom_structure.copy(), sigma=0.1)

    def test_rattle_returns_different_structure_on_multiple_calls(self):
        """Test that rattle returns distinct values on subsequent calls with same input."""
        positions1 = rattle(self.structure.copy(), sigma=1).positions
        positions2 = rattle(self.structure.copy(), sigma=1).positions
        self.assertTrue(not np.array_equal(positions1, positions2))

    def test_stretch_modifies_cell(self):
        """Test that stretch modifies the cell of the structure."""
        original_cell = self.structure.get_cell().copy()
        stretched_structure = stretch(self.structure.copy(), hydro=0.1, shear=0.1)
        stretched_cell = stretched_structure.get_cell()
        self.assertFalse(np.allclose(original_cell, stretched_cell))

    def test_rattle_class(self):
        """Test the Rattle class."""
        rattle_pert = Rattle(sigma=0.1)
        rattled_structure = rattle_pert(self.structure.copy())
        self.assertIn('rattle(0.1)', rattled_structure.info['perturbation'])

    def test_rattle_class_supercell(self):
        """Test the Rattle class with create_supercells=True."""
        rattle_pert = Rattle(sigma=0.1, create_supercells=True)
        rattled_structure = rattle_pert(self.single_atom_structure.copy())
        self.assertEqual(len(rattled_structure), 8)

    def test_stretch_class(self):
        """Test the Stretch class."""
        stretch_pert = Stretch(hydro=0.1, shear=0.1)
        stretched_structure = stretch_pert(self.structure.copy())
        self.assertIn('stretch(hydro=0.1, shear=0.1)', stretched_structure.info['perturbation'])

    def test_series_class(self):
        """Test the Series class."""
        rattle_pert = Rattle(sigma=0.1)
        stretch_pert = Stretch(hydro=0.1, shear=0.1)
        series_pert = Series((rattle_pert, stretch_pert))
        perturbed_structure = series_pert(self.structure.copy())
        self.assertIn('rattle(0.1)+stretch(hydro=0.1, shear=0.1)', perturbed_structure.info['perturbation'])

    def test_series_addition(self):
        """Test the addition of perturbations."""
        rattle_pert = Rattle(sigma=0.1)
        stretch_pert = Stretch(hydro=0.1, shear=0.1)
        series_pert = rattle_pert + stretch_pert
        perturbed_structure = series_pert(self.structure.copy())
        self.assertIn('rattle(0.1)+stretch(hydro=0.1, shear=0.1)', perturbed_structure.info['perturbation'])

    def test_random_choice_class(self):
        """Test the RandomChoice class."""
        original_rand = np.random.rand

        def side_effect(*args, **kwargs):
            if not args and not kwargs:
                # This is the call from RandomChoice: np.random.rand()
                return side_effect.choice
            # This is a call from rattle or stretch, e.g. np.random.rand(3, 3)
            return original_rand(*args, **kwargs)

        rattle_pert = Rattle(sigma=0.1)
        stretch_pert = Stretch(hydro=0.1, shear=0.1)
        # choice_a is rattle, choice_b is stretch
        random_pert = RandomChoice(rattle_pert, stretch_pert, chance=0.5)

        with patch('numpy.random.rand', side_effect=side_effect):
            # With rand()=0.8, 0.8 > 0.5 is true, so choice_a (rattle) should be called.
            side_effect.choice = 0.8
            perturbed_structure_A = random_pert(self.structure.copy())
            self.assertIn('rattle(0.1)', perturbed_structure_A.info['perturbation'])
            self.assertNotIn('stretch', perturbed_structure_A.info['perturbation'])

            # With rand()=0.2, 0.2 > 0.5 is false, so choice_b (stretch) should be called.
            side_effect.choice = 0.2
            perturbed_structure_B = random_pert(self.structure.copy())
            self.assertIn('stretch(hydro=0.1, shear=0.1)', perturbed_structure_B.info['perturbation'])
            self.assertNotIn('rattle', perturbed_structure_B.info['perturbation'])

    def test_apply_perturbations(self):
        """Test the apply_perturbations function."""
        structures = [self.structure.copy() for _ in range(3)]
        perturbations = [Rattle(sigma=0.1), Stretch(hydro=0.1, shear=0.1)]

        perturbed_structures = list(apply_perturbations(structures, perturbations))

        self.assertEqual(len(perturbed_structures), 6)  # 3 structures * 2 perturbations

    def test_apply_perturbations_with_filter(self):
        """Test the apply_perturbations function with a filter."""
        structures = [self.structure.copy()]
        perturbations = [Rattle(sigma=0.1)]

        # This filter should always return False
        false_filter = lambda s: False

        perturbed_structures = list(apply_perturbations(structures, perturbations, filters=false_filter))
        self.assertEqual(len(perturbed_structures), 0)

    def test_apply_perturbations_value_error(self):
        """Test that apply_perturbations handles ValueError."""
        structures = [self.single_atom_structure.copy()]
        perturbations = [Rattle(sigma=0.1, create_supercells=False)]  # This will raise ValueError

        # The ValueError from Rattle should be caught, and no structures should be yielded.
        perturbed_structures = list(apply_perturbations(structures, perturbations))
        self.assertEqual(len(perturbed_structures), 0)

    def test_stretch_strain_distribution(self):
        """Test that stretch applies strain within the correct ranges."""
        for _ in range(10):  # Repeat test 10 times with different random values
            hydro = np.random.uniform(0.05, 0.3)
            shear = np.random.uniform(0.05, 0.3)
            minimum_strain = np.random.uniform(1e-4, 1e-2)

            # Ensure hydro and shear are greater than minimum_strain
            hydro = max(hydro, minimum_strain + 1e-3)
            shear = max(shear, minimum_strain + 1e-3)

            original_cell = self.structure.get_cell().copy()
            stretched_structure = stretch(self.structure.copy(), hydro=hydro, shear=shear, minimum_strain=minimum_strain)
            stretched_cell = stretched_structure.get_cell()

            strain_modifier = np.linalg.inv(original_cell) @ stretched_cell

            epsilon = strain_modifier - np.identity(3)

            self.assertTrue(np.allclose(epsilon, epsilon.T), "Strain matrix should be symmetric")

            # Check diagonal elements
            diag_strains = np.diag(epsilon)
            for strain_val in diag_strains:
                self.assertTrue(minimum_strain <= abs(strain_val) <= hydro)

            # Check off-diagonal elements
            off_diag_indices = np.triu_indices(3, k=1)
            off_diag_strains = epsilon[off_diag_indices]
            for strain_val in off_diag_strains:
                self.assertTrue(minimum_strain <= abs(strain_val) <= shear)

    def test_perturbation_info_concatenation(self):
        """Test that perturbation info is concatenated."""
        rattle_pert = Rattle(sigma=0.1)
        stretch_pert = Stretch(hydro=0.1, shear=0.1)

        structure = self.structure.copy()
        structure = rattle_pert(structure)
        structure = stretch_pert(structure)

        self.assertIn('rattle(0.1)+stretch(hydro=0.1, shear=0.1)', structure.info['perturbation'])


@pytest.fixture
def reference_dict():
    """
    Per‑element scaling factors that will be multiplied by the global
    ``sigma`` argument inside ``element_scaled_rattle``.
    """
    return {
        "H": 0.8,   # Å
        "C": 1.2,   # Å
        "O": 1.0,   # Å
    }


def test_element_scaled_rattle_raises_value_error_for_single_atom():
    """element_scaled_rattle should raise a ValueError for a single-atom structure."""
    single_atom_structure = Atoms('H', positions=[[0, 0, 0]], cell=[10, 10, 10])
    with pytest.raises(ValueError):
        element_scaled_rattle(single_atom_structure.copy(), sigma=0.1, reference={"H": 1})


def test_element_scaled_rattle_missing_reference_raises():
    """element_scaled_rattle should raise a ValueError if reference dict does not contain all elements!"""
    structure = bulk("Fe")
    bad_reference = {"Cu": 1.0}   # No entry for Fe
    with pytest.raises(ValueError, match="No value for element Fe"):
        element_scaled_rattle(structure.copy(), sigma=0.2, reference=bad_reference)


@st.composite
def random_element_structures(draw):
    """Return structures with random elements inside"""
    structure = bulk("Cu", cubic=True).repeat(3)
    elements = st.lists(st.sampled_from(list(atomic_numbers.keys())[1:106]),
                        min_size=len(structure), max_size=len(structure))
    structure.symbols[:] = draw(elements)
    return structure


@given(random_element_structures(), st.floats(min_value=0.01, max_value=10))
def test_element_scaled_rattle_respects_element_specific_sigma(simple_structure, sigma):
    """
    For each element present in the structure we verify two things:

    1. The *mean* displacement of every Cartesian component is (close to) zero.
    2. The *standard deviation* of the displacement matches the expected
       ``sigma * reference[element]`` value.
    """
    unique_symbols = set(simple_structure.symbols)
    reference_dict = {sym: (i+1)/len(unique_symbols) for i, sym in enumerate(unique_symbols)}
    n_repeat = 500

    displacements = np.stack([
        element_scaled_rattle(simple_structure.copy(), sigma, reference_dict).positions
        - simple_structure.positions for _ in range(n_repeat)
    ])
    displacements = displacements.transpose([1, 0, 2])

    for sym, disp in zip(simple_structure.symbols, displacements):
        element_sigma = sigma * reference_dict[sym]
        # Instead of fancy hypothesis testing, just check that mean and variance are within five times their standard
        # error, implying they should almost never fail
        assert np.allclose(disp.mean(axis=0), 0.0,
                           atol=5 * element_sigma / np.sqrt(n_repeat))
        assert np.allclose(disp.var(axis=0),  element_sigma**2,
                           atol=5 * np.sqrt(2 * sigma**4 / (n_repeat - 1)))


if __name__ == '__main__':
    unittest.main()



from ase import Atoms
from ase.cell import Cell
from assyst.filters import AspectFilter
from hypothesis import given, strategies as st
from pyxtal.lattice import generate_cellpara


@st.composite
def cells(draw):
    ltype = st.sampled_from(["monoclinic", "triclinic", "orthorhombic", "tetragonal", "hexagonal", "trigonal", "cubic"])
    volume = st.floats(min_value=5, max_value=100, allow_nan=False, allow_infinity=False)
    return Cell.fromcellpar(generate_cellpara(draw(ltype), draw(volume)))


@given(cells(), st.floats(0, exclude_min=True))
def test_aspect_filter(cell, ratio):
    filter = AspectFilter(maximum_aspect_ratio=ratio)

    structure = Atoms('Cu', cell=cell, pbc=True)
    c_over_a = max(structure.cell.lengths()) / min(structure.cell.lengths())
    assert filter(structure) == (c_over_a <= ratio), \
        "AspectFilter should filter only structures larger than given aspect ratio!"
    filter = AspectFilter(maximum_aspect_ratio=c_over_a)
    assert filter(structure), \
        "AspectFilter should not filter structures with exactly the given aspect ratio!"



"""Crystal structure generation step of ASSYST."""

from dataclasses import dataclass
from collections.abc import Sequence
from itertools import product, islice
from warnings import catch_warnings, warn
from typing import Self, Iterable, Iterator, Literal, overload, Union

from .filters import DistanceFilter

from ase import Atoms
from pyxtal import pyxtal as _pyxtal
from pyxtal.msg import Comp_CompatibilityError
from pyxtal.tolerance import Tol_matrix
from tqdm.auto import tqdm


def _get_real_spacegroup(s):
    """Use pyxtal to sniff spacegroup of generated crystals.

    The pyxtal object does not update the space group after a call to from_random, but keeps the requested one around.
    Since the generated atoms may actually be of higher symmetry, reinitialize here to make sure what we've got."""
    p = _pyxtal()
    p.from_seed(s)
    return p.group.number


def pyxtal(
    group: Union[int, list[int]],
    species: tuple[str],
    num_ions: tuple[int],
    dim: Literal[0, 1, 2, 3] = 3,
    repeat: int = 1,
    allow_exceptions: bool = True,
    **kwargs,
) -> Union[Atoms, list[dict]]:
    """
    Generate random crystal structures with PyXtal.

    `group` must be between 1 and the largest possible value for the given dimensionality:
        dim=3 => 1 - 230 (space groups)
        dim=2 => 1 -  80 (layer groups)
        dim=1 => 1 -  75 (rod groups)
        dim=0 => 1 -  58 (point groups)

    When `group` is passed as a :class:`list` of :class:`int` or `repeat>1`, generate multiple structures and return them in a :class:`list`
    of :class:`dict` containing the keys `atoms`, `symmetry` and `repeat` for the ASE structure, the symmetry group
    number and which iteration it is, respectively.

    Args:
        group (list of int, or int): the symmetry group to generate or a list of them
        species (tuple of str): which species to include, defines the stoichiometry together with `num_ions`
        num_ions (tuple of int): how many of each species to include, defines the stoichiometry together with `species`
        dim (int): dimensionality of the symmetry group, 0 is point groups, 1 is rod groups, 2 is layer groups and 3 is space groups
        repeat (int): how many random structures to generate
        allow_exceptions (bool): when generating multiple structures, silence errors when the requested stoichiometry and symmetry group are incompatible
        **kwargs: passed to `pyxtal.pyxtal` function verbatim

    Returns:
        :class:`ase.Atoms`: the generated structure, if repeat==1 and only one symmetry group is requested
        :class:`list` of :class:`dict` of all generated structures, if repeat>1 or multiple symmetry groups are requested

    Raises:
        ValueError: if `species` and `num_ions` are not of the same length
        ValueError: if stoichiometry and symmetry group are incompatible and allow_exceptions==False or only one structure is requested
    """

    if len(species) != len(num_ions):
        raise ValueError(
            "species and num_ions must be of same length, "
            f"not {species} and {num_ions}!"
        )
    stoich = "".join(f"{s}{n}" for s, n in zip(species, num_ions))

    def generate(group):
        s = _pyxtal()
        try:
            s.from_random(
                dim=dim, group=group, species=species, numIons=num_ions, **kwargs
            )
        except Comp_CompatibilityError:
            if not allow_exceptions:
                raise ValueError(
                    f"Symmetry group {group} incompatible with stoichiometry {stoich}!"
                ) from None
            else:
                return None
        s = s.to_ase()
        s.wrap(center=(0, 0, 0))
        return s

    # return a single structure
    if repeat == 1 and isinstance(group, int):
        allow_exceptions = False
        return generate(group)
    else:
        structures = []
        if isinstance(group, int):
            group = [group]
        failed_groups = []
        for g in tqdm(group, desc="Spacegroups"):
            for i in range(repeat):
                s = generate(g)
                if s is None:
                    failed_groups.append(g)
                    continue
                structures.append({
                    "atoms": s, "symmetry": g, "repeat": i,
                    "requested spacegroup": g, "spacegroup": _get_real_spacegroup(s),
                })
        if len(failed_groups) > 0:
            warn(
                f"Groups [{', '.join(map(str, failed_groups))}] could not be generated with stoichiometry {stoich}!",
                stacklevel=2,
            )
        return structures


@dataclass(eq=True, frozen=True)
class Formulas(Sequence):
    """Simple helper to generate lists of structure compositions.

    :func:`.sample_space_groups` is the intended consumer and expects an iterable of dictionaries, where each dictionary
    maps an element name to the number of atoms of this type in one structure.
    :class:`.Formulas` behaves as if it were such a tuple, but extends the basic python arithmetic operations to make
    building the list a bit simpler.

    The class can be initialized from any tuple of dictionaries.

    >>> el_manual = Formulas(({'Cu': 1}, {'Cu': 2}))

    :meth:`.range` is a helper class method that initializes `Formulas` for a single element and takes the same
    arguments as the builtin `range`, except that it skips the zero.

    >>> el = Formulas.range('Cu', 3)
    Formulas(atoms=({'Cu': 1}, {'Cu': 2}))
    >>> el == el_manual
    True

    Addition is overloaded to the addition of the underlying tuples.

    >>> Formulas.range('Cu', 1, 5) == Formulas.range('Cu', 1, 3) + Formulas.range('Cu', 3, 5)

    The bitwise or operation is akin to the inner product

    >>> Formulas.range('Cu', 3) | Formulas.range('Ag', 3)
    Formulas(atoms=({'Cu': 1, 'Ag': 1}, {'Cu': 2, 'Ag': 2}))

    >>> Formulas.range('Cu', 3) * Formulas.range('Ag', 3)
    Formulas(atoms=({'Cu': 1, 'Ag': 1}, {'Cu': 2, 'Ag': 1}, {'Cu': 1, 'Ag': 2}, {'Cu': 2, 'Ag': 2}))
    """

    atoms: tuple[dict[str, int], ...]

    @property
    def elements(self) -> set[str]:
        """Set of elements present in elements."""
        e: set[str] = set()
        for s in self.atoms:
            e = e.union(s.keys())
        return e

    @classmethod
    def range(cls, elements: str | Iterable[str], *range_args) -> Self:
        """Creates formulas with number of atoms as given by the builtin `range`.

        Multiple elements are combined as the outer product."""
        if isinstance(elements, str):
            return cls(tuple({elements: i} for i in range(*range_args)))
        formulas = [cls.range(e, *range_args) for e in elements]
        total = formulas[0]
        for f in formulas[1:]:
            total *= f
        return total

    def __add__(self, other: Self) -> Self:
        """Extend underlying list of stoichiometries."""
        return type(self)(self.atoms + other.atoms)

    def __or__(self, other: Self) -> Self:
        """Inner product of underlying stoichiometries.

        Truncates to the length of the shortest of the two element sequences.
        Must not share elements with other.elements."""
        assert self.elements.isdisjoint(
            other.elements
        ), "Can only or stoichiometries of different elements!"
        s: tuple[dict[str, int], ...] = ()
        for me, you in zip(self.atoms, other.atoms):
            s += (me | you,)
        return type(self)(s)

    def __mul__(self, other: Self) -> Self:
        """Outer product of underlying stoichiometries.

        Must not share elements with other.elements."""
        assert self.elements.isdisjoint(
            other.elements
        ), "Can only multiply stoichiometries of different elements!"
        s: tuple[dict[str, int], ...] = ()
        for me, you in product(self.atoms, other.atoms):
            s += (me | you,)
        return type(self)(s)

    # Sequence Impl'
    @overload
    def __getitem__(self, index: int) -> dict[str, int]: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[dict[str, int]]: ...
    def __getitem__(self, index):
        return self.atoms[index]

    def __len__(self) -> int:
        return len(self.atoms)

    def trim(self, min_atoms: int = 1, max_atoms: int | None = None) -> Self:
        """Returns a copy of itself with formulas with lesser or more atoms than given limits removed."""
        if max_atoms is not None:
            return type(self)(
                tuple(f for f in self if min_atoms <= sum(f.values()) <= max_atoms)
            )
        else:
            return type(self)(tuple(f for f in self if min_atoms <= sum(f.values())))


def sample_space_groups(
    formulas: Formulas | Iterable[dict[str, int]],
    spacegroups: list[int] | tuple[int, ...] | Iterable[int] | None = None,
    min_atoms: int = 1,
    max_atoms: int = 10,
    max_structures: int | None = None,
    dim: Literal[0, 1, 2, 3] = 3,
    tolerance: (
        Literal["metallic", "atomic", "molecular", "vdW"] | DistanceFilter | dict
    ) = "metallic",
) -> Iterator[Atoms]:
    """
    Create symmetric random structures.

    Args:
        formulas (:class:`.Formulas` or :class:`collections.abc.Iterable` of :class:`dict` from :class:`str` to :class:`int`): :class:`list` of chemical formulas
        spacegroups (list of int): which space groups to generate
        min_atoms (int): do not generate structures smaller than this
        max_atoms (int): do not generate structures larger than this
        max_structures (int): generate at most this many structures
        dim (one of 0, 1, 2, or 3): the dimensionality of the structures to generate; if lower than 3 the code generates
            samples no longer from space groups, but from the subperiodic layer, rod, or point groups.
        tolerance (str, dict of elements to radii):
            specifies minimum allowed distances between atoms in generated structures;
            if str then it should be one values understood by :class:`pyxtal.tolerance.Tol_matrix`;
            if dict each value gives the minimum *radius* allowed for an atom, whether a given distance is allowed then
            depends on the sum of the radii of the respective elements

    Yields:
        :class:`ase.Atoms`: random symmetric crystal structures
    """

    if not 0 <= dim <= 3:
        raise ValueError(f"dim must be in range [0, 3], not {dim}!")

    # number of (sub-)periodic symmetry groups available in 0-3 dimensions
    max_group = [58, 75, 80, 230][dim]

    if spacegroups is None:
        spacegroups = range(1, max_group + 1)
    spacegroups = list(spacegroups)

    min_spg = min(spacegroups)
    max_spg = max(spacegroups)
    if min_spg <= 0 or max_group < max_spg:
        raise ValueError(
            f"spacegroups must be in range [1, {max_group}], not [{min_spg}, {max_spg}] (dim={dim})!"
        )

    tm: Tol_matrix | None
    match tolerance:
        case "metallic" | "atomic" | "molecular" | "vdW":
            tm = Tol_matrix(prototype=tolerance)
        case dict():
            tm = (
                DistanceFilter(tolerance).to_tol_matrix()
                if len(tolerance) > 0
                else None
            )
        case DistanceFilter():
            tm = tolerance.to_tol_matrix()
        case _:
            raise ValueError("invalid value tolerance={tolerance}!")

    for stoich in (bar := tqdm(formulas)):
        # pyxtal never returns structures when one element with zero atoms is present, so filter here first for
        # robustness
        stoich = {e: n for e, n in stoich.items() if n > 0}
        if len(stoich) == 0:
            continue
        elements, num_atoms = zip(*stoich.items())
        if not min_atoms <= sum(num_atoms) <= max_atoms:
            continue
        stoich_str = "".join(f"{s}{n}" for s, n in zip(elements, num_atoms))
        bar.set_description(stoich_str)

        def pop(s):
            atoms = s.pop("atoms")
            atoms.info.update(s)
            return atoms

        with catch_warnings(category=UserWarning, action="ignore"):
            px = pyxtal(spacegroups, elements, num_atoms, dim=dim, tm=tm)
            yield from islice(map(pop, px), max_structures)
            if max_structures is not None:
                max_structures -= len(px)
                if max_structures <= 0:
                    break


__all__ = [
        "pyxtal",
        "Formulas",
        "sample_space_groups",
]



import unittest
from assyst.calculators import Morse, Grace
from ase.calculators.morse import MorsePotential
from ase.calculators.calculator import Calculator

try:
    import tensorpotential
except ImportError:
    tensorpotential = None


class TestCalculators(unittest.TestCase):
    def test_morse(self):
        config = Morse(epsilon=2.0, r0=2.0, rho0=2.0)
        calc = config.get_calculator()
        self.assertIsInstance(calc, MorsePotential)
        self.assertEqual(calc.parameters['epsilon'], 2.0)
        self.assertEqual(calc.parameters['r0'], 2.0)
        self.assertEqual(calc.parameters['rho0'], 2.0)

    @unittest.skipIf(tensorpotential is None, "tensorpotential not installed")
    def test_grace(self):
        config = Grace()
        calc = config.get_calculator()
        self.assertIsInstance(calc, Calculator)



import unittest
from unittest.mock import patch
from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator
from assyst.relax import Relax, CellRelax, VolumeRelax, SymmetryRelax, FullRelax, relax
from assyst.calculators import AseCalculatorConfig

class MockCalculator:
    def get_potential_energy(self, atoms=None):
        return 0.0
    def get_forces(self, atoms=None):
        return [[0.0, 0.0, 0.0]]
    def get_stress(self, atoms=None):
        return [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

class TestRelax(unittest.TestCase):
    def setUp(self):
        self.structure = Atoms('H', positions=[[0, 0, 0]], cell=[10, 10, 10], pbc=True)
        self.structure.calc = MockCalculator()

    @patch('ase.optimize.LBFGS.run')
    def test_relax_runs(self, mock_run):
        relaxer = Relax()
        relaxed_structure = relaxer.relax(self.structure)
        mock_run.assert_called_once()
        self.assertIsInstance(relaxed_structure.calc, SinglePointCalculator)

    @patch('ase.optimize.LBFGS.run')
    def test_cell_relax(self, mock_run):
        relaxer = CellRelax()
        relaxer.relax(self.structure)
        mock_run.assert_called_once()

    @patch('ase.optimize.LBFGS.run')
    def test_volume_relax(self, mock_run):
        relaxer = VolumeRelax(pressure=1.0)
        relaxer.relax(self.structure)
        mock_run.assert_called_once()

    @patch('ase.optimize.LBFGS.run')
    def test_symmetry_relax(self, mock_run):
        relaxer = SymmetryRelax()
        relaxer.relax(self.structure)
        mock_run.assert_called_once()

    @patch('ase.optimize.LBFGS.run')
    def test_full_relax(self, mock_run):
        relaxer = FullRelax(pressure=1.0)
        relaxer.relax(self.structure)
        mock_run.assert_called_once()

    @patch('assyst.relax.Relax.relax')
    def test_relax_function_with_calc_object(self, mock_relax_method):
        settings = Relax()
        calculator = MockCalculator()
        structures = [self.structure]

        list(relax(settings, calculator, structures))

        self.assertIs(mock_relax_method.call_args[0][0].calc, calculator)

    @patch('assyst.relax.Relax.relax')
    def test_relax_function_with_calc_config(self, mock_relax_method):
        class MockCalcConfig(AseCalculatorConfig):
            def get_calculator(self):
                return MockCalculator()

        settings = Relax()
        calculator_config = MockCalcConfig()
        structures = [self.structure]

        list(relax(settings, calculator_config, structures))

        self.assertIsInstance(mock_relax_method.call_args[0][0].calc, MockCalculator)



import unittest
import numpy as np
from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator
from assyst.filters import EnergyFilter, ForceFilter, CalculatorFilter

class MockCalculatorFilter(CalculatorFilter):
    def __call__(self, structure: Atoms) -> bool:
        return self._check(structure)

class TestCalculatorFilters(unittest.TestCase):
    def test_calculator_filter_check(self):
        filter = MockCalculatorFilter(missing='error')

        # No calculator
        structure = Atoms('Cu')
        with self.assertRaises(ValueError):
            filter._check(structure)

        # Wrong calculator
        structure.calc = "dummy"
        with self.assertRaises(ValueError):
            filter._check(structure)

        # Correct calculator
        structure.calc = SinglePointCalculator(structure, energy=0.0, forces=np.zeros((1, 3)))
        self.assertTrue(filter._check(structure))

        # Missing ignore
        filter = MockCalculatorFilter(missing='ignore')
        structure = Atoms('Cu')
        self.assertFalse(filter(structure))

    def test_energy_filter(self):
        # Test with no calculator
        filter = EnergyFilter(max_energy=1.0, missing='ignore')
        structure = Atoms('Cu')
        self.assertTrue(filter(structure)) # Should pass if _check is false

        filter = EnergyFilter(max_energy=1.0)
        structure = Atoms('Cu')
        structure.calc = SinglePointCalculator(structure, energy=2.0)
        self.assertFalse(filter(structure)) # energy 2.0 > 1.0

        structure.calc = SinglePointCalculator(structure, energy=0.5)
        self.assertTrue(filter(structure)) # energy 0.5 < 1.0

        filter = EnergyFilter(min_energy=0.0, max_energy=1.0)
        structure.calc = SinglePointCalculator(structure, energy=-0.5)
        self.assertFalse(filter(structure)) # energy -0.5 < 0.0

    def test_force_filter(self):
        # Test with no calculator
        filter = ForceFilter(max_force=1.0, missing='ignore')
        structure = Atoms('Cu')
        self.assertTrue(filter(structure)) # Should pass if _check is false

        filter = ForceFilter(max_force=1.0)
        structure = Atoms('Cu')
        forces = np.array([[1.1, 0.0, 0.0]])
        structure.calc = SinglePointCalculator(structure, forces=forces)
        self.assertFalse(filter(structure)) # force 1.1 > 1.0

        forces = np.array([[0.5, 0.0, 0.0]])
        structure.calc = SinglePointCalculator(structure, forces=forces)
        self.assertTrue(filter(structure)) # force 0.5 < 1.0



"""Classes to apply (random) perturbations to structures."""

from abc import ABC, abstractmethod
from ase import Atoms
from typing import Iterable, Callable, Self, Iterator
from dataclasses import dataclass
import numpy as np

from .filters import Filter


def rattle(structure: Atoms, sigma: float) -> Atoms:
    """Randomly displace positions with gaussian noise.

    Operates INPLACE."""
    if len(structure) == 1:
        raise ValueError("Can only rattle structures larger than one atom.")
    structure.rattle(stdev=sigma, rng=np.random)
    return structure


def element_scaled_rattle(structure: Atoms, sigma: float, reference: dict[str, float]) -> Atoms:
    """Randomly displace positions with gaussian noise relative to an elemental reference length.

    Operates like :func:`.rattle` but uses a standard deviation derived from the relative `sigma` and the `reference`,
    where this reference is given by element.

    Operates IN PLACE!

    Args:
        structure (:class:`ase.Atoms`): structure to perturb
        sigma (:class:`float`): relative standard deviation
        reference (:class:`dict` of :class:`str` to :class:`float`): reference length per element

    Raises:
        ValueError: if len(structure) == 1, create a super cell first before calling again
        ValueError: if reference values are not positive
        ValueError: if reference does not contain all elements in given structure
    """
    sigma = sigma * np.ones(len(structure))
    if not all(r > 0 for r in reference.values()):
        raise ValueError("Reference lengths must be strictly positive!")
    for i, sym in enumerate(structure.symbols):
        try:
            sigma[i] *= reference[sym]
        except KeyError:
            raise ValueError(f"No value for element {sym} provided in argument `reference`!") from None
    return rattle(structure, sigma.reshape(-1, 1))


def stretch(structure: Atoms, hydro: float, shear: float, minimum_strain=1e-3) -> Atoms:
    """Randomly stretch cell with uniform noise.

    Ensures at least `minimum_strain` strain to avoid structures very close to their original structures.
    These don't offer a lot of new information and can also confuse VASP's symmetry analyzer.

    Operates INPLACE."""

    def get_strains(max_strain, size):
        signs = np.random.choice([-1, 1], size=size)
        magnitudes = np.random.uniform(minimum_strain, max_strain, size=size)
        return signs * magnitudes

    strain = np.zeros((3, 3))
    # Off-diagonal elements
    indices = np.triu_indices(3, k=1)
    strain[indices] = get_strains(shear, 3)
    strain += strain.T

    # Diagonal elements
    np.fill_diagonal(strain, 1 + get_strains(hydro, 3))

    structure.set_cell(structure.cell.array @ strain, scale_atoms=True)
    return structure


class PerturbationABC(ABC):
    """Apply some perturbation to a given structure."""

    def __call__(self, structure: Atoms) -> Atoms:
        if "perturbation" not in structure.info:
            structure.info["perturbation"] = str(self)
        else:
            structure.info["perturbation"] += "+" + str(self)
        return structure

    @abstractmethod
    def __str__(self) -> str:
        pass

    def __add__(self, other: Self) -> "Series":
        return Series((self, other))


Perturbation = Callable[[Atoms], Atoms] | PerturbationABC


def apply_perturbations(
    structures: Iterable[Atoms],
    perturbations: Iterable[Perturbation],
    filters: Iterable[Filter] | Filter | None = None,
    retries: int = 10,
) -> Iterator[Atoms]:
    """Apply a list of perturbations to each structure and yield the result of each perturbation separately.

    If a perturbation raises ValueError it is ignored.

    Args:
        structures: :class:`collections.abc.Iterable` of :class:`ase.Atoms` to perturb.
        perturbations: :class:`collections.abc.Iterable` of :class:`~.Perturbation` that modify structures.
        filters: :class:`collections.abc.Iterable` of :class:`~assyst.filters.Filter` to filter valid results (optional).
        retries: :class:`int`, max attempts per perturbation (default: 10).

    Yields:
        :class:`ase.Atoms`: perturbed structure that passes all filters.
    """
    if filters is None:
        filters = []
    if not isinstance(filters, Iterable):
        filters = [filters]
    perturbations = list(perturbations)

    for structure in structures:
        for mod in perturbations:
            try:
                for _ in range(retries):
                    m = mod(structure.copy())
                    if all(f(m) for f in filters):
                        yield m
                        break
            except ValueError:
                continue


@dataclass(frozen=True)
class Rattle(PerturbationABC):
    """Displace atoms by some absolute amount from a normal distribution."""

    sigma: float
    create_supercells: bool = False
    "Create minimal 2x2x2 super cells when applied to structures of only one atom."

    def __call__(self, structure: Atoms):
        if self.create_supercells and len(structure) == 1:
            structure = structure.repeat(2)
        structure = super().__call__(structure)
        return rattle(structure, self.sigma)

    def __str__(self):
        return f"rattle({self.sigma})"


@dataclass(frozen=True)
class ElementScaledRattle(PerturbationABC):
    """Displace atoms by some amount from a normal distribution.

    Operates like :class:`.Rattle` but uses a standard deviation derived from the relative `sigma` and the `reference`,
    where this reference is given by element.
    """

    sigma: float
    reference: dict[str, float]
    create_supercells: bool = False
    "Create minimal 2x2x2 super cells when applied to structures of only one atom."

    def __call__(self, structure: Atoms):
        if self.create_supercells and len(structure) == 1:
            structure = structure.repeat(2)
        structure = super().__call__(structure)
        return element_scaled_rattle(structure, self.sigma, self.reference)

    def __str__(self):
        return f"scaled_rattle({self.sigma})"


@dataclass(frozen=True)
class Stretch(PerturbationABC):
    """Apply random cell perturbation."""

    hydro: float
    shear: float
    minimum_strain: float = 1e-3

    def __call__(self, structure: Atoms):
        structure = super().__call__(structure)
        return stretch(structure, self.hydro, self.shear, self.minimum_strain)

    def __str__(self):
        return f"stretch(hydro={self.hydro}, shear={self.shear})"


@dataclass(frozen=True)
class Series(PerturbationABC):
    """Apply some perturbations in sequence."""

    perturbations: tuple[Perturbation, ...]

    def __call__(self, structure: Atoms) -> Atoms:
        for mod in self.perturbations:
            structure = mod(structure)
        return structure

    def __str__(self):
        return "+".join(str(mod) for mod in self.perturbations)


@dataclass(frozen=True)
class RandomChoice(PerturbationABC):
    """Apply either of two alternatives randomly."""

    choice_a: Perturbation
    choice_b: Perturbation
    chance: float
    "Probability to pick choice b"

    def __call__(self, structure: Atoms) -> Atoms:
        if np.random.rand() > self.chance:
            return self.choice_a(structure)
        else:
            return self.choice_b(structure)

    def __str__(self):
        return str(self.choice_a) + "|" + str(self.choice_b)


__all__ = [
        "rattle",
        "stretch",
        "PerturbationABC",
        "Perturbation",
        "apply_perturbations",
        "Rattle",
        "Stretch",
        "Series",
        "RandomChoice",
]



from unittest.mock import Mock
from ase import Atoms
from assyst.perturbations import apply_perturbations

def test_filter_retry_success():
    """Test that apply_perturbations retries and eventually succeeds when filter fails."""
    structure = Atoms('H2', positions=[[0, 0, 0], [0.74, 0, 0]], cell=[10, 10, 10])
    # Filter fails 3 times, then succeeds.
    ff = Mock(side_effect=[False, False, False, True])
    identity = lambda s: s
    results = list(apply_perturbations([structure], [identity], filters=[ff], retries=10))
    assert len(results) == 1
    assert ff.call_count == 4

def test_filter_retry_failure():
    """Test that apply_perturbations gives up after retries limit when filter keeps failing."""
    structure = Atoms('H2', positions=[[0, 0, 0], [0.74, 0, 0]], cell=[10, 10, 10])
    # Filter always fails
    ff = Mock(return_value=False)
    identity = lambda s: s
    results = list(apply_perturbations([structure], [identity], filters=[ff], retries=10))
    assert len(results) == 0
    assert ff.call_count == 10

def test_value_error_no_retry():
    """Test that apply_perturbations does NOT retry and silently ignores when ValueError is raised."""
    structure = Atoms('H2', positions=[[0, 0, 0], [0.74, 0, 0]], cell=[10, 10, 10])
    fp = Mock(side_effect=ValueError("Intentional ValueError"))

    results = list(apply_perturbations([structure], [fp], retries=10))
    assert len(results) == 0
    assert fp.call_count == 1

def test_mixed_no_retry_on_value_error():
    """Test that ValueError stops retries even if some attempts succeeded in reaching filters."""
    structure = Atoms('H2', positions=[[0, 0, 0], [0.74, 0, 0]], cell=[10, 10, 10])
    # First call returns structure, second raises ValueError
    fp = Mock(side_effect=[structure, ValueError("Intentional ValueError")])
    # Filter always fails
    ff = Mock(return_value=False)

    results = list(apply_perturbations([structure], [fp], filters=[ff], retries=10))
    assert len(results) == 0
    assert fp.call_count == 2



from ase import Atoms
from ase.cell import Cell
from assyst.filters import VolumeFilter

# TODO unify duplicated code
from hypothesis import given, strategies as st
from pyxtal.lattice import generate_cellpara


@st.composite
def cells(draw):
    ltype = st.sampled_from(["monoclinic", "triclinic", "orthorhombic", "tetragonal", "hexagonal", "trigonal", "cubic"])
    volume = st.floats(min_value=5, max_value=100, allow_nan=False, allow_infinity=False)
    return Cell.fromcellpar(generate_cellpara(draw(ltype), draw(volume)))


@given(cells(), st.floats(0, exclude_min=True))
def test_volume_filter(cell, maximum_volume_per_atom):
    filter = VolumeFilter(maximum_volume_per_atom)

    structure = Atoms('Cu', cell=cell, pbc=True)
    volume = structure.cell.volume/len(structure)
    assert filter(structure) == (volume <= maximum_volume_per_atom), \
        "VolumeFilter should filter only structures larger than given volume!"
    filter = VolumeFilter(volume)
    assert filter(structure), \
        "VolumeFilter should not filter structures with exactly the given volume!"



import unittest
from unittest.mock import patch, MagicMock
import numpy as np
from ase import Atoms
from assyst.plot import (
    _volume,
    _energy,
    _concentration,
    volume_histogram,
    size_histogram,
    concentration_histogram,
    distance_histogram,
    radial_distribution,
    energy_histogram,
    energy_volume,
)

try:
    import matscipy
except ImportError:
    matscipy = None


class TestPlotHelpers(unittest.TestCase):
    def setUp(self):
        self.s1 = Atoms('H2', positions=[[0,0,0], [1,0,0]], cell=[10,10,10])
        self.s2 = Atoms('O', positions=[[0,0,0]], cell=[5,5,5])
        self.s1.calc = MagicMock()
        self.s1.calc.get_potential_energy.return_value = -2.0
        self.s2.calc = MagicMock()
        self.s2.calc.get_potential_energy.return_value = -5.0
        self.structures = [self.s1, self.s2]

    def test_volume(self):
        volumes = _volume(self.structures)
        self.assertAlmostEqual(volumes[0], 500.0)
        self.assertAlmostEqual(volumes[1], 125.0)

    def test_energy(self):
        energies = _energy(self.structures)
        self.assertAlmostEqual(energies[0], -1.0)
        self.assertAlmostEqual(energies[1], -5.0)

    def test_concentration(self):
        concentrations = _concentration(self.structures)
        self.assertTrue('H' in concentrations)
        self.assertTrue('O' in concentrations)
        np.testing.assert_array_almost_equal(concentrations['H'], [1.0, 0.0])
        np.testing.assert_array_almost_equal(concentrations['O'], [0.0, 1.0])

    def test_concentration_with_elements(self):
        concentrations = _concentration(self.structures, elements=['H'])
        self.assertTrue('H' in concentrations)
        self.assertFalse('O' in concentrations)

class TestPlotFunctions(unittest.TestCase):
    def setUp(self):
        self.structures = [Atoms('H2O', positions=[[0,0,0], [1,0,0], [0,1,0]], cell=[10,10,10])]

    @patch('matplotlib.pyplot.hist')
    def test_volume_histogram(self, mock_hist):
        volume_histogram(self.structures)
        mock_hist.assert_called_once()

    @patch('matplotlib.pyplot.hist')
    def test_size_histogram(self, mock_hist):
        size_histogram(self.structures)
        mock_hist.assert_called_once()

    @patch('matplotlib.pyplot.bar')
    def test_concentration_histogram(self, mock_bar):
        concentration_histogram(self.structures)
        mock_bar.assert_called()

    @unittest.skipIf(matscipy is None, "matscipy not installed")
    @patch('matplotlib.pyplot.hist')
    def test_distance_histogram(self, mock_hist):
        distance_histogram(self.structures, reduce="min")
        mock_hist.assert_called_once()

        distance_histogram(self.structures, reduce="mean")
        distance_histogram(self.structures, reduce=None)

    @unittest.skipIf(matscipy is None, "matscipy not installed")
    @patch('matplotlib.pyplot.hist')
    def test_radial_distribution(self, mock_hist):
        radial_distribution(self.structures)
        mock_hist.assert_called_once()

    @patch('matplotlib.pyplot.hist')
    def test_energy_histogram(self, mock_hist):
        s = Atoms('H', cell=[10,10,10])
        s.calc = MagicMock()
        s.calc.get_potential_energy.return_value = -1.0
        energy_histogram([s])
        mock_hist.assert_called_once()

    @patch('matplotlib.pyplot.scatter')
    def test_energy_volume_scatter(self, mock_scatter):
        s = Atoms('H', cell=[10,10,10])
        s.calc = MagicMock()
        s.calc.get_potential_energy.return_value = -1.0
        energy_volume([s])
        mock_scatter.assert_called_once()

    @patch('matplotlib.pyplot.hexbin')
    def test_energy_volume_hexbin(self, mock_hexbin):
        s = Atoms('H', cell=[10,10,10])
        s.calc = MagicMock()
        s.calc.get_potential_energy.return_value = -1.0
        energy_volume([s] * 1001)
        mock_hexbin.assert_called_once()



"""Helper plotting functions."""

from typing import Literal, Callable, Iterable
from collections import Counter, defaultdict

from ase import Atoms
import matplotlib.pyplot as plt
import numpy as np

from assyst.neighbors import neighbor_list


def _volume(structures: Iterable[Atoms]) -> list[float]:
    return [s.cell.volume / len(s) for s in structures]


def _energy(structures: Iterable[Atoms]) -> list[float]:
    return [s.get_potential_energy() / len(s) for s in structures]


def _concentration(
    structures: Iterable[Atoms], elements: Iterable[str] | None = None
) -> list[dict[str, float]]:
    structure_concentrations = [
        {k: v / len(s) for k, v in Counter(s.symbols).items()} for s in structures
    ]
    concentrations = defaultdict(lambda: np.zeros(len(structure_concentrations)))
    for i, d in enumerate(structure_concentrations):
        for e, c in d.items():
            concentrations[e][i] = c
    if elements is not None:
        concentrations = {e: concentrations[e] for e in elements}
    return concentrations


def _distance(
    structures: Iterable[Atoms], rmax: float
) -> list[Iterable[float]]:
    return [neighbor_list("d", s, float(rmax)) for s in structures]


def volume_histogram(structures: list[Atoms], **kwargs):
    """Plot histogram of per-atom volumes.

    Args:
        structures (list of :class:`ase.Atoms`):
            structures to plot
        **kwargs:
            passed through to :func:`matplotlib.pyplot.hist`

    Returns:

        Return value of :func:`matplotlib.pyplot.hist`"""
    res = plt.hist(_volume(structures), **kwargs)
    plt.xlabel(r"Volume [$\mathrm{\AA}^3/\mathrm{atom}$]")
    plt.ylabel(r"#$\,$Structures")
    return res


def size_histogram(structures: list[Atoms], **kwargs):
    """Plot histogram of number of atoms.

    Args:
        structures (list of :class:`ase.Atoms`):
            structures to plot
        **kwargs:
            passed through to :func:`matplotlib.pyplot.hist`

    Returns:
        Return value of :func:`matplotlib.pyplot.hist`"""
    res = plt.hist(list(map(len, structures)), **kwargs)
    plt.xlabel("# Atoms")
    plt.ylabel(r"#$\,$Structures")
    return res


def concentration_histogram(
    structures: list[Atoms], elements: Iterable[str] | None = None, **kwargs
):
    """Plot histogram of concentrations.

    Args:
        structures (list of :class:`ase.Atoms`):
            structures to plot
        elements (iterable of str):
            which element concentrations to plot, by default all present
        **kwargs:
            passed through to :func:`matplotlib.pyplot.bar`"""
    conc = _concentration(structures, elements=elements)
    conc_step = np.diff(
        sorted(np.unique(np.concatenate([np.unique(c) for c in conc.values()])))
    ).min()
    kwargs.setdefault("width", conc_step)
    width = kwargs["width"]
    kwargs["width"] = width / len(conc)
    shifts = np.linspace(0, 1, len(conc), endpoint=False)
    for i, (e, c) in enumerate(conc.items()):
        x, h = np.unique(c, return_counts=True)
        plt.bar(x + shifts[i] * width - width / 2, h, label=e, align="edge", **kwargs)
    plt.legend()
    plt.xlabel("Concentration")
    plt.ylabel("#$\\,$Structures")


def distance_histogram(
    structures: list[Atoms],
    rmax: float = 6.0,
    reduce: Literal["min", "mean"] | Callable[[Iterable[float]], float] | None = "min",
    **kwargs,
):
    """Plot histogram of per-atom volumes.

    Args:
        structures (list of :class:`ase.Atoms`):
            structures to plot
        rmax (float):
            maximum cutoff to consider neighborhood
        reduce (callable from array of floats to float):
            applied to the neighbor distances per structure, and should reduce a single scalar that is binned
        **kwargs:
            passed through to :func:`matplotlib.pyplot.hist`

    Returns:
        Return value of :func:`matplotlib.pyplot.hist`"""
    kwargs.setdefault("bins", 100)
    labels = {
        "min": r"Minimum distance [$\mathrm{\AA}$]",
        "mean": r"Mean distance [$\mathrm{\AA}$]",
    }
    xlabel = labels.get(reduce, r"Distance [$\mathrm{\AA}$]")

    _preset = {
        "min": np.min,
        "mean": np.mean,
    }

    if reduce is None:
        data = np.concatenate(
            [neighbor_list("d", s, float(rmax)) for s in structures]
        )
        ylabel = r"#$\,$Neighbours"
    else:
        reduce_func = _preset.get(reduce, reduce)
        data = [reduce_func(neighbor_list("d", s, float(rmax))) for s in structures]
        ylabel = r"#$\,$Structures"

    res = plt.hist(data, **kwargs)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    return res


def radial_distribution(
    structures: list[Atoms],
    rmax: float = 6.0,
    **kwargs
):
    """Plot radial distribution of neighbors in training set.

    Calculates all neighbors in all structures and histograms them together.
    Bins are weighted by 1/(4 pi r^2), but because the density in each
    structure can be different, the plot does *not* yield something that can be
    directly compared to a Radial Distribution Function.
    It can be used to locate prefered bonding distances or sampling of the
    radial neighborhood in a training set given suitable data.

    Args:
        structures (list of :class:`ase.Atoms`):
            structures to plot
        rmax (float):
            maximum cutoff to consider neighborhood
        **kwargs: pass through to :func:`matplotlib.pyplot.hist`

    Returns:
        Return value of :func:`matplotlib.pyplot.hist`"""
    kwargs.setdefault("bins", 100)
    distances = np.concatenate([n for n in _distance(structures, rmax)])
    weights = 1 / (4 * np.pi * distances ** 2)
    res = plt.hist(distances, weights=weights, **kwargs)
    plt.xlabel(r"Distance [$\mathrm{\AA}$]")
    plt.ylabel("Radial distribution")
    return res


def energy_histogram(
        structures: list[Atoms],
        **kwargs
):
    """Plot energy per atom histogram.

    Requires that :class:`ase.calculators.singlepoint.SinglePointCalculator` are attached to the atoms, either from a
    relaxation for final training set calculation.

    Args:
        structures (list of :class:`ase.Atoms`): structures to plot
        **kwargs: pass through to :func:`matplotlib.pyplot.hist`

    Returns:
        Return value of :func:`matplotlib.pyplot.hist`"""
    kwargs.setdefault("bins", 100)
    E = _energy(structures)
    res = plt.hist(E, **kwargs)
    plt.xlabel(r"Energy [eV/atom]")
    plt.ylabel(r"#$\,$Structures")
    return res


def energy_volume(structures: list[Atoms], **kwargs):
    """Plot energy per atom versus volume per atom.

    Requires that :class:`ase.calculators.singlepoint.SinglePointCalculator` are attached to the atoms, either from a
    relaxation for final training set calculation.

    Args:
        structure: list[Atoms],
            structures to plot"""
    V = _volume(structures)
    E = _energy(structures)
    structures = list(structures)
    if len(structures) < 1000:
        if "s" not in kwargs and "markersize" not in kwargs:
            kwargs["markersize"] = 5
        plt.scatter(V, E, **kwargs)
    else:
        plt.hexbin(V, E, **kwargs, bins="log")
    plt.xlabel(r"Volume [$\mathrm{\AA}^3/\mathrm{atom}$]")
    plt.ylabel(r"Energy [eV/atom]")


__all__ = [
        "volume_histogram",
        "size_histogram",
        "concentration_histogram",
        "distance_histogram",
        "radial_distribution",
        "energy_histogram",
        "energy_volume",
]
